/** This class only contains all the variables of the NES Picture Processing Unit, also known as PPU. <BR>
    This code is GPL. :o)
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
public class NESPictureProcessingUnit{

    /** Video RAM, also known as VRAM. <BR>
	0x4000 bytes of RAM. */
    protected char Memory[] = new char[ 0x4000 ];

    /** Sprite RAM. <BR>
	0x100 bytes of RAM. */
    protected char SpriteRam[] = new char[ 0x100 ];

    /** Indicates if NMI can be executed during VBlank. <BR>
	True: execute NMI on VBlank.<BR>
	False: do not execute NMI on VBlank. */
    protected boolean IsExecuteNMIOnVBlankEnabled;

    /** Indicates the PPU configuration. <BR>
	True: PPU is slave.<BR>
	False: PPU is master.<BR>
	This feature is unused by the NES. */
    protected boolean IsPPUSlave;

    /** Indicates sprite size which can be 8 or 16. */
    protected byte SpriteSize;

    /** Indicates background pattern table address base which can be 0 or 0x100 (for cache implementation purpose). */
    protected char BackgroundPatternTableAddress;
    
    /** Indicates sprite pattern table address base which can be 0 or 0x100 (for cache implementation purpose). */
    protected char SpritePatternTableAddress;

    /** Indicates which value should be added to VRAM address: 1 or 32. */
    protected byte VRAMAddressIncrement;

    /** Indicates which display type is used: color or monochrome. */
    protected boolean IsDisplayTypeColorDisplay;

    /** Indicates if background is visible in the left 8-pixel column. <BR>
	Values will be used to optimize code. */
    protected int BackgroundClipping;

    /** Indicates if sprites are visible in the left 8-pixel column. <BR>
	Values will be used to optimize code. */
    protected int SpriteClipping;

    /** Indicates if background is visible. <BR>
	True: background is visible.<BR>
	False: background is invisible. */
    protected boolean BackgroundVisibility;

    /** Indicates if sprites are visible. <BR>
	True: sprites are visible.<BR>
	False: sprites are invisible. */
    protected boolean SpriteVisibility;

    /** Indicates which color component (R, G, B) should be intensified. <BR>
	Values will be used to optimize code. <BR>
	This feature is incomplete. */
    protected byte ColorIntensity;

    /** Indicates background color for monochrome display mode. <BR>
	Values will be used to optimize code. <BR>
	This feature is incomplete. */
    protected byte FullBackgroundColor;

    /** Indicates if byte is first or second half address. <BR>
	True: first write (high byte).<BR>
	False: second write (low byte).<BR>
	Mainly used to control writes to 0x2005 and 0x2006 (they share this indicator). */
    protected boolean AddressMode;

    /** Indicates if sprite 0 was hit. <BR>
	True: sprite 0 has been hit.<BR>
	False: sprite 0 has not been hit.<BR>
	This feature is incomplete. */
    protected boolean WasSpriteZeroHit = false;

    /* Indicates if VBlank is occuring. <BR>
       True: VBlank is occuring.<BR>
       False: VBlank is not occuring. */
    //protected boolean IsVBlankOccuring;

    /* Indicates if there are more than 8 sprites on current scanline. <BR>
       True: There are more than 8 sprites on current sanline.<BR>
       False: there are less than 9 sprites on current scanline.<BR>
       This feature is incomplete. */
    protected boolean IsThereMoreThanEightSpritesOnScanline = false;

    /* Indicates if VRAM writes are respected. <BR>
       True: VRAM writes are respected.<BR>
       False: VRAM writes are ignored.<BR>
       I must check this, but i do not know how to change this value.*/
    //protected boolean AreVRAMWritesRespected;

    /** Buffer used to access VRAM data (except for color palette). */
    protected char VRAMDataBuffer;

    /** First half of the VRAM address (high byte). */
    protected char HalfVRAMAddress;

    /** Complete VRAM address (word). */
    protected char VRAMAddress;

    /** VRAM address register ($2006). */
    protected char VRAMAddressRegister;

    /** Temp VRam Address register. <BR>
	xYYYXXXXXXXXXXXX where<BR>
	XXXXXXXXXXXX is nametable address (-$2000)<BR>
	YYY is tile Y offset<BR>
	x is unsued. */
    protected char TempVRAMAddressRegister;

    /** Used in scroll effects. <BR>
	Control the horizontal scroll (0 to 7 pixels). */
    protected int XOffset;

    /** Current scanline. <BR>
	Possible values are 0 to 255. */
    protected char Scanline;

    /** Indicates nameTable 1 address according to mirroring (vertical or horizontal, specified in the NES file header). <BR>
	Value can be 0x2000 or 0x2400, current implementation is imcomplete. */
    protected char NameTable1;

    /** Indicates nameTable 2 address according to mirroring (vertical or horizontal, specified in the NES file header). <BR>
	Value can be 0x2000 or 0x2400, current implementation is incomplete. */
    protected char NameTable2;

    /** Constructor. */
    public void NESPictureProcessingUnit(){}

}
