/** The NMOS6502 class emulates the standard NMOS 6502 microprocessor. <BR>
    This is an early version that has not been completely tested !!<BR>
    So it is buggy. ;o)<BR>
    Unofficial opcodes and addressing modes are not implemented.<BR>
    IRQ implementation is incomplete.
    I would like to thank all the people who helped (huge list).<BR>
    This code is GPL. :o)<BR>
    This version is fast enough.<BR>
    It can be extended to other CPU quite easily (you will have to modify this code a little bit if you plan to port Juergen's code, i will try to explain it later, or maybe i will do it).<BR>
    Why not port Mame/Mess to Java ??<BR>
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    <B>Based on :</B><BR>
    m6502<BR>
    Portable 6502/65c02/65sc02/6510/n2a03 emulator V1.2<BR>
    Copyright (c) 1998 Juergen Buchmueller, all rights reserved.<BR>
    65sc02 core Copyright (c) 2000 Peter Trauner.<DL>
    <UR>
    <LI> This source code is released as freeware for non-commercial purposes.<BR>
    <LI> You are free to use and redistribute this code in modified or unmodified form, provided you list me in the credits.<BR>
    <LI> If you modify this source code, you must add a notice to each modified source file that it has been changed. If you're a nice person, you will clearly mark each change too.  :)<BR>
    <LI> If you wish to use this for commercial purposes, please contact me at <A HREF="mailto:pullmoll@t-online.de">pullmoll@t-online.de</A><BR>
    <LI> The author of this copywritten work reserves the right to change the terms of its usage and license at any time, including retroactively<BR>
    <LI> This entire notice must remain in the source code.
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "A good documentation on NMOS 6502 for more details."
    @see "The Gnu Public License." */
public abstract class NMOS6502{
    
    /** Previous program counter. */
    private Pair previousProgramCounter = new Pair();
    
    /** Program counter. */
    protected Pair programCounter = new Pair();
    
    /** Stack pointer. <BR>
	Always between 0x0100 and 0x01FF. */
    protected Pair stackPointer = new Pair();

    /** Zero page address. */
    private Pair zeroPageAddress = new Pair();

    /** Effective address. */
    protected Pair effectiveAddress = new Pair();

    /** Accumulator register. */
    protected UInt8 accumulator = new UInt8();

    /** X index register. */
    protected UInt8 xIndexRegister = new UInt8();

    /** Y index register. */
    protected UInt8 yIndexRegister = new UInt8();

    /** Processor status register. <BR>
	Flags in the status register are as follows:<DL>
	N V T B D I Z C<BR>
	7 6 5 4 3 2 1 0<BR><BR>
	N: Negative flag (also called S flag for sign flag).<BR>
	V: Overflow flag.<BR>
	T: Bit number 5 is unused. It is always set and it is initialized by reset() so you would better call it before running the processor.<BR>
	B: Break flag.<BR>
	D: Decimal mode status flag.<BR>
	I: Interrupt flag.<BR>
	Z: Zero flag.<BR>
	C: Carry flag.*/
    protected UInt8 processorStatus = new UInt8();

    /** Pending IRQ and last instruction cleared I. */
    private int after_cli;

    /** NMI state. */
    private int nmi_state = 0;

    /** IRQ state. */
    private int irq_state = 0;
    
    /** Set overflow state. */
    private int so_state = 0;

    /** Nonzero if a NMI or IRQ is pending. <BR>
	If pendingInterrupt = 0, no interrupt pending.<BR>
	If pendingInterrupt = 1, an IRQ is pending.<BR>
	If pendingInterrupt = 2, a NMI is pending.*/
    protected int pending_irq = 0;

    /** IRQ callbak. */
    private int irq_callback = 0;
    
    /** Number of cycles left. */
    protected int iCount = 0;

    /** Temporary variable used in adressing modes. */
    protected int tmp = 0;

    /** No interrupt value. <BR>
	Replace a #define M6502_INT_NONE 0. */
    private final int M6502_INT_NONE = 0;

    /** IRQ interrupt value. <BR>
	Replace a #define M6502_INT_IRQ 1. */
    private final int M6502_INT_IRQ = 1;
    
    /** NMI interrupt value. <BR>
	Replace a #define M6502_INT_NMI 2. */
    private final int M6502_INT_NMI = 2;

    /** SET_OVERFLOW value. <BR>
	Replace a #define M6502_SET_OVERFLOW 3. Use cpu_set_irq_line(cpu, M6502_SET_OVERFLOW, level) to change level of the so input line positiv edge sets overflow flag. */
    private final int M6502_SET_OVERFLOW = 3;

    /** NMI vector value. <BR>
	Replace a #define M6502_NMI_VEC 0xFFFA. */
    private final int M6502_NMI_VEC = 0xFFFA;

    /** Reset vector value. <BR>
	Replace a #define M6502_RST_VEC 0xFFFC.*/
    private final int M6502_RST_VEC = 0xFFFC;

    /** IRQ vector value. <BR>
	Replace a #define M6502_IRQ_VEC 0xFFFE. */
    private final int M6502_IRQ_VEC = 0xFFFE;

    /** C flag value. <BR>
	Replace a #define F_C 0x01. */
    protected final int F_C = 0x01;

    /** Z flag value. <BR>
	Replace a #define F_Z 0x02. */
    protected final int F_Z = 0x02;

    /** I flag value. <BR>
	Replace a #define F_I 0x04. */
    private final int F_I = 0x04;

    /** D flag value. <BR>
	Replace a #define F_D 0x08. */
    protected final int F_D = 0x08;

    /** B flag value. <BR>
	Replace a #define F_B 0x10. */
    private final int F_B = 0x10;

    /** T flag value. <BR>
	Replace a #define F_T 0x20. */
    private final int F_T = 0x20;

    /** V flag value. <BR>
	Replace a #define F_V 0x40. */
    protected final int F_V = 0x40;

    /** N flag value. <BR>
	Replace a #define F_N 0x80. */
    protected final int F_N = 0x80;

    /** Clear line value. <BR>
	Replace a #define CLEAR_LINE 0. */
    private final int CLEAR_LINE = 0;

    /** Constructor. */
    public NMOS6502(){}

    /** Display the different registers. <BR>
	It produces the following output:<BR><BR>
	Accumulator: xxx   X register: xxx    Y register: xxx<BR>
	Stack pointer: xxx    Program counter: xxxx    Status flags: NV5BDIZC<BR><BR>
	Accumulator, X and Y registers are integer values between 0 and 255.<BR>
	Stack pointer is an integer value between 0 and 255 but you have to add 0x0100 to get the true value.<BR>
	Program counter is an integer value between 0 and 65535.<BR>
	Why not improve it and display hexadecimal values?<BR>
	Status register display use the following rules:<BR>
	<UR>
	<LI> 5 is always displayed (since bit 5 is always set).<BR>
	<LI> If a bit is set, the letter corresponding to the flag is displayed. Otherwise (i.e. bit is cleared), an underscore ("_") is displayed.
	</UR> */
    final protected void debugRegisters(){
	System.out.println( "Accumulator: " + (int) accumulator.get() + "\tX register: " + (int) xIndexRegister.get() + "\tY register: " + (int) yIndexRegister.get() );
	System.out.print( "Stack pointer: " + (int) stackPointer.getLowByte() + "\tProgram counter: "+(int)programCounter.getWord() + "\tStatus flags: ");
	if( ( processorStatus.get() & 0x80) == 0x80 ) System.out.print( "N" );
	else System.out.print( "_" );
	if( ( processorStatus.get() & 0x40 ) == 0x40 ) System.out.print( "V" );
	else System.out.print( "_" );
	System.out.print( "5" );
	if( ( processorStatus.get() & 0x20 ) == 0x10 ) System.out.print( "B" );
	else System.out.print( "_" );
	if( ( processorStatus.get() & 0x08 ) == 0x08 ) System.out.print( "D" );
	else System.out.print( "_" );
	if( ( processorStatus.get() & 0x04 ) == 0x04 ) System.out.print( "I" );
	else System.out.print( "_" );
	if( ( processorStatus.get() & 0x02 ) == 0x02 ) System.out.print( "Z" );
	else System.out.print( "_" );
	if( ( processorStatus.get() & 0x01 ) == 0x01 ) System.out.println( "C" );
	else System.out.println( "_" );
    }
    
    /** This is provided as a dummy function. <BR>
	Some system will use it when program counter change (if you plan to use it you will have to modify this source by quitting the final and private attributes and make it protected and abstract for example).
	@param value is used for the new program counter. */
    final private void changePC16( int value ){}
    
    /** Write to memory. <BR>
	You have to provide this function.
	@param address the memory address where you want to write.
	@param data the value you want to write. */
    abstract protected void WRMEM( int address, int data );

    /** Read from memory. <BR>
	You have to provide this function.
	@param address is the memory address you want to read from.
	@return the unsigned byte read in memory. */
    abstract protected int RDMEM( int address );
    
    /** Set N and Z flags in the processor status register according to condition.
	@param condition The condition.<BR>
	If condition = 0, N is cleared and Z is set.<BR>
	Else N is set if it is set in condition and Z is cleared.
    */
    final protected void setNZ( int condition ){
	if ( condition == 0 ) processorStatus.set( ( processorStatus.get() & ~F_N ) | F_Z );
	else processorStatus.set( ( processorStatus.get() & ~( F_N | F_Z ) ) | ( condition & F_N ) );
    }
    
    /** Read an opcode.
	@return the opcode value. */
    final private int readOpcode(){
	final int temp;
	temp = RDMEM( programCounter.getWord() );
	programCounter.setWord( programCounter.getWord() + 1 );
	return( temp );
    }
    
    /** Read an opcode argument.
	@return the opcode value.*/
    final private int readOpcodeArgument(){
	final int temporary;
	temporary = RDMEM( programCounter.getWord() );
	programCounter.setWord( programCounter.getWord() + 1 );
	return( temporary );
    }
    
    /** Branch relative. <BR>
	An extra cycle is used if page boundary is crossed.
	@param condition that indicates if the branch relative occurs.*/
    final private void BRA( boolean condition ){       
	if( condition )
	    {		
		tmp = readOpcodeArgument();
		effectiveAddress.setWord( programCounter.getWord() + ( byte ) tmp );
		iCount -= ( programCounter.getHighByte() == effectiveAddress.getHighByte() ) ? 3 : 4;
		programCounter.setDWord( effectiveAddress.getDWord() );
		changePC16( programCounter.getDWord() );
	    }
	else
	    {			
		programCounter.setWord( programCounter.getWord() + 1 );
		iCount -= 2;
	    }
    }
    
    // Helper macros to build the effective address

    /** Effective address = zero page address. */
    final private void EA_ZPG(){
	zeroPageAddress.setLowByte( readOpcodeArgument() );
	effectiveAddress.setDWord( zeroPageAddress.getDWord() );
    }

    /** Effective address = zero page address + X. */
    final private void EA_ZPX(){
	zeroPageAddress.setLowByte( readOpcodeArgument() + xIndexRegister.get() );
	effectiveAddress.setDWord( zeroPageAddress.getDWord() );
    }

    /** Effective address = zero page address + Y. */
    final private void EA_ZPY(){
	zeroPageAddress.setLowByte( readOpcodeArgument() + yIndexRegister.get() );
	effectiveAddress.setDWord( zeroPageAddress.getDWord() );
    }

    /** Effective address = absolute address. */
    final private void EA_ABS(){
	effectiveAddress.setLowByte( readOpcodeArgument() );
	effectiveAddress.setHighByte( readOpcodeArgument() );
    }

    /** Effective address = absolute address + X. */
    final protected void EA_ABX(){
	EA_ABS();
	effectiveAddress.setWord( effectiveAddress.getWord() + xIndexRegister.get() );
    }

    /** Effective address = absolute address + Y. */
    final protected void EA_ABY(){
	EA_ABS();
	effectiveAddress.setWord( effectiveAddress.getWord() + yIndexRegister.get() );
    }

    /** Effective address = zero page + X indirect. <BR>
	This is called pre indexed mode. */
    final private void EA_IDX(){
	zeroPageAddress.setLowByte( readOpcodeArgument() + xIndexRegister.get() );
	effectiveAddress.setLowByte( RDMEM( zeroPageAddress.getDWord() ) );
	zeroPageAddress.setLowByte( zeroPageAddress.getLowByte() + 1 );
	effectiveAddress.setHighByte( RDMEM( zeroPageAddress.getDWord() ) );
    }

    /** Effective address = zero page indirect + Y (post indexed). <BR>
	Subtract 1 cycle if page boundary is crossed. */
    final private void EA_IDY(){
	zeroPageAddress.setLowByte( readOpcodeArgument() );
	effectiveAddress.setLowByte( RDMEM( zeroPageAddress.getDWord() ) );
	zeroPageAddress.setLowByte( zeroPageAddress.getLowByte() + 1 );
	effectiveAddress.setHighByte( RDMEM( zeroPageAddress.getDWord() ) );
	if ( ( effectiveAddress.getLowByte() + yIndexRegister.get() ) > 0xFF ) iCount--;
	effectiveAddress.setWord( effectiveAddress.getWord() + yIndexRegister.get() );
    }

    /** Effective address = indirect. <BR>
	This is only only used by JMP. */
    final private void EA_IND(){
	EA_ABS();
	tmp = RDMEM( effectiveAddress.getDWord() );
	effectiveAddress.setLowByte( effectiveAddress.getLowByte() + 1 );	/* booby trap: stay in same page! ;-) */
	effectiveAddress.setHighByte( RDMEM( effectiveAddress.getDWord() ) );
	effectiveAddress.setLowByte( tmp );
    }

    /** Read directly a value and store it into tmp. */
    final protected void RD_IMM(){
	tmp = readOpcodeArgument();
    }

    /** Read accumulator value and store it into tmp. */
    final private void RD_ACC(){
	tmp = accumulator.get();
    }

    /** Read a value using EA_ZPG and store it into tmp. */
    final protected void RD_ZPG(){
	EA_ZPG();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_ZPX and store it into tmp. */
    final protected void  RD_ZPX(){
	EA_ZPX();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_ZPY and store it into tmp. */
    final protected void RD_ZPY(){
	EA_ZPY();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_ABS and store it into tmp. */
    final protected void RD_ABS(){
	EA_ABS();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }
    
    /** Read a value using EA_ABX and store it into tmp. */
    final protected void RD_ABX(){
	EA_ABX();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_ABY and store it into tmp. */
    final protected void RD_ABY(){
	EA_ABY();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_IDX and store it into tmp. */
    final protected void RD_IDX(){
	EA_IDX();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Read a value using EA_IDY and store it into tmp. */
    final protected void RD_IDY(){
	EA_IDY();
	tmp = RDMEM( effectiveAddress.getDWord() );
    }

    /** Write a value from tmp using EA_ZPG. */
    final protected void WR_ZPG(){
	EA_ZPG();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_ZPX. */
    final protected void WR_ZPX(){
	EA_ZPX();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_ZPY. */
    final protected void WR_ZPY(){
	EA_ZPY();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_ABS. */
    final protected void WR_ABS(){
	EA_ABS();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_ABX. */
    final protected void WR_ABX(){
	EA_ABX();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_ABY. */
    final protected void WR_ABY(){
	EA_ABY();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_IDX. */
    final protected void WR_IDX(){
	EA_IDX();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /** Write a value from tmp using EA_IDY. */
    final protected void WR_IDY(){
	EA_IDY();
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    /* Write back a value from tmp to accumulator. */
    final protected void WB_ACC(){
	accumulator.set( tmp );
    }

    /* Write back a value from tmp to the last efective address. */
    final protected void WB_EA(){
	WRMEM( effectiveAddress.getDWord(), tmp);
    }

    // Macros to emulate the plain 6502 opcodes

    /** Push a register onto the stack.
	@param register the byte to push.*/
    final private void PUSH( int register ){
	WRMEM(stackPointer.getDWord(), register);
	stackPointer.setLowByte( stackPointer.getLowByte() - 1 );
    }

    /** Pull a register from the stack.
	@return the byte pulled.*/
    final private int PULL(){
	stackPointer.setLowByte( stackPointer.getLowByte() + 1 );
	return( RDMEM( stackPointer.getDWord() ) );
    }

    /** Add with carry. */
    protected void ADC(){
	if ( ( processorStatus.get() & F_D ) != 0 )
	    {
		final int c = ( processorStatus.get() & F_C );
		int lo = ( accumulator.get() & 0x0F ) + ( tmp & 0x0F ) + c;
		int hi = ( accumulator.get() & 0xF0 ) + ( tmp & 0xF0 );
		processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
		if ( lo > 0x09 )
		    {
			hi += 0x10;
			lo += 0x06;
		    }
		if ( ( ~ ( accumulator.get() ^ tmp ) & ( accumulator.get() ^ hi ) & F_N ) != 0 ) processorStatus.set( processorStatus.get() | F_V );
		if ( hi > 0x90 ) hi += 0x60;
		if ( ( hi & 0xFF00 ) != 0 ) processorStatus.set( processorStatus.get() | F_C );
		accumulator.set( ( lo & 0x0F ) + ( hi & 0xF0 ) );
	    }
	else
	    {
		final int c = ( processorStatus.get() & F_C );
		final int sum = accumulator.get() + tmp + c;
		processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
		if ( ( ~ ( accumulator.get() ^ tmp ) & ( accumulator.get() ^ sum ) & F_N ) != 0 ) processorStatus.set( processorStatus.get() | F_V );
		if ( ( sum & 0xFF00 ) != 0 ) processorStatus.set( processorStatus.get() | F_C );
		accumulator.set( sum );
	    }
	setNZ( accumulator.get() );
    }

    /** Logical AND. */
    final private void AND(){
	accumulator.set( accumulator.get() & tmp );
	setNZ( accumulator.get() );
    }

    /** Arithmetic shift left. */
    final private void ASL(){
	processorStatus.set( ( processorStatus.get() & ~F_C ) | ( ( tmp >>> 7 ) & F_C ) );
	tmp = ( tmp << 1 ) & 0x00FF;
	setNZ( tmp );
    }

    /** Branch if carry clear. */
    final private void BCC(){
	BRA( ( processorStatus.get() & F_C ) == 0 );
    }

    /** Branch if carry set. */
    final private void BCS(){
	BRA( ( processorStatus.get() & F_C ) != 0 );
    }

    /** Branch if equal. */
    final private void BEQ(){
	BRA( ( processorStatus.get() & F_Z ) != 0 );
    }

    /** Bit test. */
    final private void BIT(){
	processorStatus.set( processorStatus.get() & ~( F_N | F_V | F_Z ) );
	processorStatus.set( processorStatus.get() | ( tmp & ( F_N | F_V ) ) );
	if ( ( tmp & accumulator.get() ) == 0 ) processorStatus.set( processorStatus.get() | F_Z );
    }

    /** Branch if minus. */
    final private void BMI(){
	BRA( ( processorStatus.get() & F_N ) != 0 );
    }

    /** Branch if not equal. */
    final private void BNE(){
	BRA( ( processorStatus.get() & F_Z ) == 0 );
    }

    /** Branch if plus. */
    final private void BPL(){
	BRA( ( processorStatus.get() & F_N ) == 0 );
    }

    /** Break. <BR>
	Increment PC, push PC high byte, PC low byte, flags (with B bit set), set I flag, reset D flag and jump via IRQ vector. */
    final private void BRK(){
	programCounter.setWord( programCounter.getWord() + 1 );
	PUSH( programCounter.getHighByte() );
	PUSH( programCounter.getLowByte() );
	PUSH( processorStatus.get() | F_B );
	processorStatus.set( ( processorStatus.get() | F_I ) & ~F_D );
	programCounter.setLowByte( RDMEM( M6502_IRQ_VEC ) );
	programCounter.setHighByte( RDMEM( M6502_IRQ_VEC + 1 ) );
	changePC16( programCounter.getDWord() );
    }

    /** Branch if overflow clear. */
    final private void BVC(){
	BRA( ( processorStatus.get() & F_V ) == 0 );
    }

    /** Branch if overflow set. */
    final private void BVS(){
	BRA( ( processorStatus.get() & F_V ) != 0 );
    }

    /** Clear carry flag. */
    final private void CLC(){
	processorStatus.set( processorStatus.get() & ~F_C );
    }
    
    /** Clear decimal flag. */
    final private void CLD(){
	processorStatus.set( processorStatus.get() & ~F_D );
    }
    
    /** Clear interrupt flag. */
    final private void CLI(){
	if ( ( irq_state != CLEAR_LINE ) && ( ( processorStatus.get() & F_I ) != 0 ) )
	    {
		after_cli = 1;
	    }
	processorStatus.set( processorStatus.get() & ~F_I );
    }
    
    /** Clear overflow flag. */
    final private void CLV(){
	processorStatus.set( processorStatus.get() & ~F_V );
    }
    
    /** Compare accumulator. */
    final private void CMP(){
	processorStatus.set( processorStatus.get() & ~F_C );
	if ( accumulator.get() >= tmp ) processorStatus.set( processorStatus.get() | F_C );
	setNZ( accumulator.get() - tmp );
    }
    
    /** Compare index X. */
    final private void CPX(){
	processorStatus.set( processorStatus.get() & ~F_C );
	if ( xIndexRegister.get() >= tmp ) processorStatus.set( processorStatus.get() | F_C );
	setNZ( xIndexRegister.get() - tmp );
    }
    
    /** Compare index Y. */
    final private void CPY(){
	processorStatus.set( processorStatus.get() & ~F_C );
	if ( yIndexRegister.get() >= tmp ) processorStatus.set( processorStatus.get() | F_C );
	setNZ( yIndexRegister.get() - tmp );
    }
    
    /** Decrement memory. */
    final private void DEC(){
	tmp = ( -- tmp ) & 0x00FF;
	setNZ( tmp );
    }
    
    /** Decrement index X. */
    final private void DEX(){
	xIndexRegister.set( xIndexRegister.get() - 1 );
	setNZ( xIndexRegister.get() );
    }

    /** Decrement index Y. */
    final private void DEY(){
	yIndexRegister.set( yIndexRegister.get() - 1 );
	setNZ( yIndexRegister.get() );
    }

    /** Logical exclusive OR. */
    final private void EOR(){
	accumulator.set( accumulator.get() ^ tmp );
	setNZ( accumulator.get() );
    }

    /** Illegal opcode. */
    final private void ILL(){}

    /** Increment memory. */
    final private void INC(){
	tmp = ( ++ tmp ) & 0x00FF;
	setNZ( tmp );
    }

    /** Increment index X. */
    final private void INX(){
	xIndexRegister.set( xIndexRegister.get() + 1 );
	setNZ( xIndexRegister.get() );
    }

    /** Increment index Y. */
    final private void INY(){
	yIndexRegister.set( yIndexRegister.get() + 1 );
	setNZ( yIndexRegister.get() );
    }

    /** Jump to address. <BR>
	Set PC to the effective address. */
    final private void JMP(){
	if( effectiveAddress.getDWord() == previousProgramCounter.getDWord() && !( pending_irq != 0 ) && !( after_cli != 0) )
	    if( iCount > 0 ) iCount = 0;
	programCounter.setDWord( effectiveAddress.getDWord() );
	changePC16( programCounter.getDWord() );
    }

    /** Jump to subroutine. <BR>
	Decrement PC (sic!), push PC high byte, push PC low byte and set
	PC to the effective address. */
    final private void JSR(){
	effectiveAddress.setLowByte( readOpcodeArgument() );
	PUSH( programCounter.getHighByte() );
	PUSH( programCounter.getLowByte() );
	effectiveAddress.setHighByte( readOpcodeArgument() );
	programCounter.setDWord( effectiveAddress.getDWord() );
	changePC16( programCounter.getDWord() );
    }

    /** Load accumulator. */
    final private void LDA(){
	accumulator.set( tmp );
	setNZ( accumulator.get() );
    }

    /** Load index X. */
    final private void LDX(){
	xIndexRegister.set( tmp );
	setNZ( xIndexRegister.get() );
    }

    /** Load index Y. */
    final private void LDY(){
	yIndexRegister.set( tmp );
	setNZ( yIndexRegister.get() );
    }

    /** Logic shift right. <BR><BR>
	0 -> [7][6][5][4][3][2][1][0] -> C */
    final protected void LSR(){
	processorStatus.set( ( processorStatus.get() & ~F_C ) | ( tmp & F_C ) );
	tmp = ( tmp >>> 1 ) & 0x00FF;
	setNZ( tmp );
    }

    /** No operation. */
    final protected void NOP(){}

    /** Logical inclusive OR. */
    final private void ORA(){
	accumulator.set( accumulator.get() | tmp );
	setNZ( accumulator.get() );
    }

    /** Push accumulator. */
    final private void PHA(){
	PUSH( accumulator.get() );
    }

    /** Push processor status (flags). */
    final private void PHP(){
	PUSH( processorStatus.get() );
    }

    /** Pull accumulator. */
    final private void PLA(){
	accumulator.set( PULL() );
	setNZ( accumulator.get() );
    }

    /** Pull processor status (flags). */
    final private void PLP(){
	if ( ( processorStatus.get() & F_I ) != 0 ) {
	    processorStatus.set( PULL() );
	    if ( ( irq_state != CLEAR_LINE ) && ( ( processorStatus.get() & F_I ) == 0 ) ) after_cli = 1;
	}
	else processorStatus.set( PULL() );
	processorStatus.set( processorStatus.get() | F_T );
    }
    
    /** Rotate left. <BR><BR>
	new C <- [7][6][5][4][3][2][1][0] <- C */
    final private void ROL(){
	tmp = ( tmp << 1 ) | ( processorStatus.get() & F_C );
	processorStatus.set( ( processorStatus.get() & ~F_C ) | ( ( tmp >>> 8 ) & F_C ) );
	tmp &= 0x00FF;
	setNZ( tmp );
    }
    
    /** Rotate right. <BR><BR>
	C -> [7][6][5][4][3][2][1][0] -> new C */
    final protected void ROR(){
	tmp |= ( processorStatus.get() & F_C ) << 8;
	processorStatus.set( ( processorStatus.get() & ~F_C ) | ( tmp & F_C ) );
	tmp = ( tmp >>> 1 ) & 0x00FF;
	setNZ( tmp );
    }
    
    /** Return from interrupt. <BR>
	Pull flags, pull PC low byte, pull PC high byte. */
    final private void RTI(){
	processorStatus.set( PULL() );
	programCounter.setLowByte( PULL() );
	programCounter.setHighByte( PULL() );
	processorStatus.set( processorStatus.get() | F_T );
	if( ( irq_state != CLEAR_LINE ) && ( ( processorStatus.get() & F_I ) == 0 ) ) after_cli = 1;
	changePC16( programCounter.getDWord() );
    }
    
    /** Return from subroutine. <BR><BR>
	Pull PC low byte, PC high byte and increment PC. */
    final private void RTS(){
	programCounter.setLowByte( PULL() );
	programCounter.setHighByte( PULL() );
	programCounter.setWord( programCounter.getWord() + 1 );
	changePC16( programCounter.getDWord() );
    }
    
    /** Subtract with carry. */
    protected void SBC(){
	if ( ( processorStatus.get() & F_D ) != 0 )
	    {
		final int c = ( processorStatus.get() & F_C ) ^ F_C;
		final int sum = accumulator.get() - tmp - c;
		int lo = ( accumulator.get() & 0x0F ) - ( tmp & 0x0F ) - c;
		int hi = ( accumulator.get() & 0xF0 ) - ( tmp & 0xF0 );
		processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
		if ( ( ( accumulator.get() ^ tmp ) & ( accumulator.get() ^ sum ) & F_N ) != 0 ) processorStatus.set( processorStatus.get() | F_V );
		if ( ( lo & 0xF0 ) != 0 ) lo -= 6;
		if ( ( lo & 0x80 ) != 0 ) hi -= 0x10;
		if ( ( hi & 0x0F00 ) != 0 ) hi -= 0x60;
		if ( ( sum & 0xFF00 ) == 0 ) processorStatus.set( processorStatus.get() | F_C );
		accumulator.set( ( lo & 0x0F ) + ( hi & 0xF0 ) );
	    }
	else
	    {
		final int c = ( processorStatus.get() & F_C ) ^ F_C;
		final int sum = accumulator.get() - tmp - c;
		processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
		if ( ( ( accumulator.get() ^ tmp ) & ( accumulator.get() ^ sum ) & F_N ) != 0 ) processorStatus.set( processorStatus.get() | F_V );
		if ( ( sum & 0xFF00 ) == 0 ) processorStatus.set( processorStatus.get() | F_C );
		accumulator.set( sum );
	    }
	setNZ( accumulator.get() );
    }
    
    /** Set carry flag. */
    final private void SEC(){
	processorStatus.set( processorStatus.get() | F_C );
    }
    
    /** Set decimal flag. */
    final private void SED(){
	processorStatus.set( processorStatus.get() | F_D );
    }
    
    /** Set interrupt flag. */
    final private void SEI(){
	processorStatus.set( processorStatus.get() | F_I );
    }
    
    /** Store accumulator. */
    final private void STA(){
	tmp = accumulator.get();
    }
    
    /** Store index X. */
    final private void STX(){
	tmp = xIndexRegister.get();
    }
    
    /** Store index Y. */
    final private void STY(){
	tmp = yIndexRegister.get();
    }
    
    /** Transfer accumulator to index X. */
    final private void TAX(){
	xIndexRegister.set( accumulator .get() );
	setNZ( xIndexRegister.get() );
    }
    
    /** Transfer accumulator to index Y. */
    final private void TAY(){
	yIndexRegister.set( accumulator .get() );
	setNZ( yIndexRegister.get() );
    }
    
    /** Transfer stack LSB to index X. */
    final private void TSX(){
	xIndexRegister.set( stackPointer.getLowByte() );
	setNZ( xIndexRegister.get() );
    }
    
    /** Transfer index X to accumulator. */
    final private void TXA(){
	accumulator.set( xIndexRegister.get() );
	setNZ( accumulator.get() );
    }
    
    /** Transfer index X to stack LSB. <BR>
	No flags changed (sic!). */
    final private void TXS(){
	stackPointer.setLowByte( xIndexRegister.get() );
    }
    
    /** Transfer index Y to accumulator. */
    final private void TYA(){
	accumulator.set( yIndexRegister.get() );
	setNZ( accumulator.get() );
    }

    /** Opcode 0x80. <BR>
	2 ILL. */
    private void opcode0x80(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x02. <BR>
	2 ILL. */
    private void opcode0x02(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x22. <BR>
	2 ILL. */
    private void opcode0x22(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x42. <BR>
	2 ILL. */
    private void opcode0x42(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x62. <BR>
	2 ILL. */
    private void opcode0x62(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x82. <BR>
	2 ILL. */
    private void opcode0x82(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0xC2. <BR>
	2 ILL. */
    private void opcode0xC2(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0xE2. <BR>
	2 ILL. */
    private void opcode0xE2(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x12. <BR>
	2 ILL. */
    private void opcode0x12(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x32. <BR>
	2 ILL. */
    private void opcode0x32(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x52. <BR>
	2 ILL. */
    private void opcode0x52(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x72. <BR>
	2 ILL. */
    private void opcode0x72(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x92. <BR>
	2 ILL. */
    private void opcode0x92(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xB2. <BR>
	2 ILL. */
    private void opcode0xB2(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xD2. <BR>
	2 ILL. */
    private void opcode0xD2(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0xF2. <BR>
	2 ILL. */
    private void opcode0xF2(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x03. <BR>
	2 ILL. */
    private void opcode0x03(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x23. <BR>
	2 ILL. */
    private void opcode0x23(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x43. <BR>
	2 ILL. */
    private void opcode0x43(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x63. <BR>
	2 ILL. */
    private void opcode0x63(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x83. <BR>
	2 ILL. */
    private void opcode0x83(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xA3. <BR>
	2 ILL. */
    private void opcode0xA3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xC3. <BR>
	2 ILL. */
    private void opcode0xC3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xE3. <BR>
	2 ILL. */
    private void opcode0xE3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x13. <BR>
	2 ILL. */
    private void opcode0x13(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x33. <BR>
	2 ILL. */
    private void opcode0x33(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x53. <BR>
	2 ILL. */
    private void opcode0x53(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x73. <BR>
	2 ILL. */
    private void opcode0x73(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x93. <BR>
	2 ILL. */
    private void opcode0x93(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xB3. <BR>
	2 ILL. */
    private void opcode0xB3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xD3. <BR>
	2 ILL. */
    private void opcode0xD3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xF3. <BR>
	2 ILL. */
    private void opcode0xF3(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x04. <BR>
	2 ILL. */
    private void opcode0x04(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x44. <BR>
	2 ILL. */
    private void opcode0x44(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    

    /** Opcode 0x64. <BR>
	2 ILL. */
    private void opcode0x64(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x14. <BR>
	2 ILL. */
    private void opcode0x14(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x34. <BR>
	2 ILL. */
    private void opcode0x34(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    

    /** Opcode 0x54. <BR>
	2 ILL. */
    private void opcode0x54(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x74. <BR>
	2 ILL. */
    private void opcode0x74(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0xD4. <BR>
	2 ILL. */
    private void opcode0xD4(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    

    /** Opcode 0xF4. <BR>
	2 ILL. */
    private void opcode0xF4(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x07. <BR>
	2 ILL. */
    private void opcode0x07(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x27. <BR>
	2 ILL. */
    private void opcode0x27(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x47. <BR>
	2 ILL. */
    private void opcode0x47(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x67. <BR>
	2 ILL. */
    private void opcode0x67(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x87. <BR>
	2 ILL. */
    private void opcode0x87(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xA7. <BR>
	2 ILL. */
    private void opcode0xA7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xC7. <BR>
	2 ILL. */
    private void opcode0xC7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xE7. <BR>
	2 ILL. */
    private void opcode0xE7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x17. <BR>
	2 ILL. */
    private void opcode0x17(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x37. <BR>
	2 ILL. */
    private void opcode0x37(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x57. <BR>
	2 ILL. */
    private void opcode0x57(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x77. <BR>
	2 ILL. */
    private void opcode0x77(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x97. <BR>
	2 ILL. */
    private void opcode0x97(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xB7. <BR>
	2 ILL. */
    private void opcode0xB7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xD7. <BR>
	2 ILL. */
    private void opcode0xD7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xF7. <BR>
	2 ILL. */
    private void opcode0xF7(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x89. <BR>
	2 ILL. */
    private void opcode0x89(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x1A. <BR>
	2 ILL. */
    private void opcode0x1A(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x3A. <BR>
	2 ILL. */
    private void opcode0x3A(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x5A. <BR>
	2 ILL. */
    private void opcode0x5A(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x7A. <BR>
	2 ILL. */
    private void opcode0x7A(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xDA. <BR>
	2 ILL. */
    private void opcode0xDA(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xFA. <BR>
	2 ILL. */
    private void opcode0xFA(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
 
    /** Opcode 0x0B. <BR>
	2 ILL. */
    private void opcode0x0B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x2B. <BR>
	2 ILL. */
    private void opcode0x2B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x4B. <BR>
	2 ILL. */
    private void opcode0x4B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x6B. <BR>
	2 ILL. */
    private void opcode0x6B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x8B. <BR>
	2 ILL. */
    private void opcode0x8B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xAB. <BR>
	2 ILL. */
    private void opcode0xAB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xCB. <BR>
	2 ILL. */
    private void opcode0xCB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xEB. <BR>
	2 ILL. */
    private void opcode0xEB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x1B. <BR>
	2 ILL. */
    private void opcode0x1B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x3B. <BR>
	2 ILL. */
    private void opcode0x3B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x5B. <BR>
	2 ILL. */
    private void opcode0x5B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x7B. <BR>
	2 ILL. */
    private void opcode0x7B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x9B. <BR>
	2 ILL. */
    private void opcode0x9B(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xBB. <BR>
	2 ILL. */
    private void opcode0xBB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xDB. <BR>
	2 ILL. */
    private void opcode0xDB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xFB. <BR>
	2 ILL. */
    private void opcode0xFB(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x0C. <BR>
	2 ILL. */
    private void opcode0x0C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x1C. <BR>
	2 ILL. */
    private void opcode0x1C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x3C. <BR>
	2 ILL. */
    private void opcode0x3C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x5C. <BR>
	2 ILL. */
    private void opcode0x5C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x7C. <BR>
	2 ILL. */
    private void opcode0x7C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x9C. <BR>
	2 ILL. */
    private void opcode0x9C(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xDC. <BR>
	2 ILL. */
    private void opcode0xDC(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x9E. <BR>
	2 ILL. */
    private void opcode0x9E(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xFC. <BR>
	2 ILL. */
    private void opcode0xFC(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x0F. <BR>
	2 ILL. */
    private void opcode0x0F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x2F. <BR>
	2 ILL. */
    private void opcode0x2F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x4F. <BR>
	2 ILL. */
    private void opcode0x4F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x6F. <BR>
	2 ILL. */
    private void opcode0x6F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x8F. <BR>
	2 ILL. */
    private void opcode0x8F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xAF. <BR>
	2 ILL. */
    private void opcode0xAF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xCF. <BR>
	2 ILL. */
    private void opcode0xCF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xEF. <BR>
	2 ILL. */
    private void opcode0xEF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x1F. <BR>
	2 ILL. */
    private void opcode0x1F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x3F. <BR>
	2 ILL. */
    private void opcode0x3F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x5F. <BR>
	2 ILL. */
    private void opcode0x5F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0x7F. <BR>
	2 ILL. */
    private void opcode0x7F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }
    
    /** Opcode 0x9F. <BR>
	2 ILL. */
    private void opcode0x9F(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xBF. <BR>
	2 ILL. */
    private void opcode0xBF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xDF. <BR>
	2 ILL. */
    private void opcode0xDF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Opcode 0xFF. <BR>
	2 ILL. */
    private void opcode0xFF(){
	iCount -= 2;		 ILL();	// 2 ILL	  
    }

    /** Switch for opcode interpretation.
	@param opcode the opcode to be interpreted. */
    final private void switchOpcode( int opcode ){
	/******************************************************************
	 ******************************************************************
	 *
	 *   plain vanilla 6502 opcodes
	 *
	 ******************************************************************
	 * op	 	  temp	cycles	rdmem		opc		wrmem   ********************/
	switch( opcode ){
	case 0x00 : {		iCount -= 7;		 BRK();		  }break; // 7 BRK
	case 0x20 : {		iCount -= 6;		 JSR();		  }break; // 6 JSR
	case 0x40 : {		iCount -= 6;		 RTI();		  }break; // 6 RTI
	case 0x60 : {		iCount -= 6;		 RTS();		  }break; // 6 RTS
	case 0x80 : { opcode0x80();                                         }break;
	case 0xA0 : { tmp = 0;	iCount -= 2; RD_IMM(); LDY();		  }break; // 2 LDY IMM
	case 0xC0 : { tmp = 0;	iCount -= 2; RD_IMM(); CPY();		  }break; // 2 CPY IMM
	case 0xE0 : { tmp = 0;	iCount -= 2; RD_IMM(); CPX();		  }break; // 2 CPX IMM
	    
	case 0x10 : { tmp = 0;							 BPL();		  }break; // 2 BPL REL
	case 0x30 : { tmp = 0;							 BMI();		  }break; // 2 BMI REL
	case 0x50 : { tmp = 0;							 BVC();		  }break; // 2 BVC REL
	case 0x70 : { tmp = 0;							 BVS();		  }break; // 2 BVS REL
	case 0x90 : { tmp = 0;							 BCC();		  }break; // 2 BCC REL
	case 0xB0 : { tmp = 0;							 BCS();		  }break; // 2 BCS REL
	case 0xD0 : { tmp = 0;							 BNE();		  }break; // 2 BNE REL
	case 0xF0 : { tmp = 0;							 BEQ();		  }break; // 2 BEQ REL
	    
	case 0x01 : { tmp = 0; iCount -= 6; RD_IDX(); ORA();		  }break; // 6 ORA IDX
	case 0x21 : { tmp = 0; iCount -= 6; RD_IDX(); AND();		  }break; // 6 AND IDX
	case 0x41 : { tmp = 0; iCount -= 6; RD_IDX(); EOR();		  }break; // 6 EOR IDX
	case 0x61 : { tmp = 0; iCount -= 6; RD_IDX(); ADC();		  }break; // 6 ADC IDX
	case 0x81 : { tmp = 0; iCount -= 6;		 STA(); WR_IDX(); }break; // 6 STA IDX
	case 0xA1 : { tmp = 0; iCount -= 6; RD_IDX(); LDA();		  }break; // 6 LDA IDX
	case 0xC1 : { tmp = 0; iCount -= 6; RD_IDX(); CMP();		  }break; // 6 CMP IDX
	case 0xE1 : { tmp = 0; iCount -= 6; RD_IDX(); SBC();		  }break; // 6 SBC IDX
	    
	case 0x11 : { tmp = 0; iCount -= 5; RD_IDY(); ORA();		  }break; // 5 ORA IDY
	case 0x31 : { tmp = 0; iCount -= 5; RD_IDY(); AND();		  }break; // 5 AND IDY
	case 0x51 : { tmp = 0; iCount -= 5; RD_IDY(); EOR();		  }break; // 5 EOR IDY
	case 0x71 : { tmp = 0; iCount -= 5; RD_IDY(); ADC();		  }break; // 5 ADC IDY
	case 0x91 : { tmp = 0; iCount -= 6;		 STA(); WR_IDY(); }break; // 6 STA IDY
	case 0xB1 : { tmp = 0; iCount -= 5; RD_IDY(); LDA();		  }break; // 5 LDA IDY
	case 0xD1 : { tmp = 0; iCount -= 5; RD_IDY(); CMP();		  }break; // 5 CMP IDY
	case 0xF1 : { tmp = 0; iCount -= 5; RD_IDY(); SBC();		  }break; // 5 SBC IDY
	    
	case 0x02 : { opcode0x02();						  }break; // 2 ILL
	case 0x22 : { opcode0x22();						  }break; // 2 ILL
	case 0x42 : { opcode0x42();						  }break; // 2 ILL
	case 0x62 : { opcode0x62();						  }break; // 2 ILL
	case 0x82 : { opcode0x82();                         		  }break;
	case 0xA2 : { tmp = 0;	iCount -= 2; RD_IMM(); LDX();		  }break; // 2 LDX IMM
	case 0xC2 : { opcode0xC2();                                 	  }break; 
	case 0xE2 : { opcode0xE2();                              	  }break; 
	    
	case 0x12 : { opcode0x12();							  }break; // 2 ILL
	case 0x32 : { opcode0x32();							  }break; // 2 ILL
	case 0x52 : { opcode0x52();							  }break; // 2 ILL
	case 0x72 : { opcode0x72();							  }break; // 2 ILL
	case 0x92 : { opcode0x92();							  }break; // 2 ILL
	case 0xB2 : { opcode0xB2();							  }break; // 2 ILL
	case 0xD2 : { opcode0xD2();							  }break; // 2 ILL
	case 0xF2 : { opcode0xF2();							  }break; // 2 ILL
	    
	case 0x03 : { opcode0x03();                         		  }break;
	case 0x23 : { opcode0x23();                         		  }break;
	case 0x43 : { opcode0x43();                                	  }break; 
	case 0x63 : { opcode0x63();                              	  }break; 
	case 0x83 : { opcode0x83();                       		  }break;
	case 0xA3 : { opcode0xA3();                              	  }break; 
	case 0xC3 : { opcode0xC3();                        		  }break; 
	case 0xE3 : { opcode0xE3();                       		  }break; 
	    
	case 0x13 : { opcode0x13();                              	  }break;
	case 0x33 : { opcode0x33();                               	  }break;
	case 0x53 : { opcode0x53();                      		  }break; 
	case 0x73 : { opcode0x73();                              	  }break; 
	case 0x93 : { opcode0x93();                              	  }break; 
	case 0xB3 : { opcode0xB3();                               	  }break; 
	case 0xD3 : { opcode0xD3();                               	  }break;
	case 0xF3 : { opcode0xF3();                      		  }break;
	    
	case 0x04 : { opcode0x04();                      		  }break;
	case 0x24 : { tmp = 0;	iCount -= 3; RD_ZPG(); BIT();		  }break; // 3 BIT ZPG
	case 0x44 : { opcode0x44();                              	  }break;
	case 0x64 : { opcode0x64();                       		  }break;
	case 0x84 : { tmp = 0;	iCount -= 3;		 STY(); WR_ZPG(); }break; // 3 STY ZPG
	case 0xA4 : { tmp = 0;	iCount -= 3; RD_ZPG(); LDY();		  }break; // 3 LDY ZPG
	case 0xC4 : { tmp = 0;	iCount -= 3; RD_ZPG(); CPY();		  }break; // 3 CPY ZPG
	case 0xE4 : { tmp = 0;	iCount -= 3; RD_ZPG(); CPX();		  }break; // 3 CPX ZPG
	    
	case 0x14 : { opcode0x14();                       		  }break;
	case 0x34 : { opcode0x34();                        		  }break; 
	case 0x54 : { opcode0x54();                              	  }break;
	case 0x74 : { opcode0x74();                       		  }break;
	case 0x94 : { tmp = 0;	iCount -= 4;		 STY(); WR_ZPX(); }break; // 4 STY ZPX
	case 0xB4 : { tmp = 0;	iCount -= 4; RD_ZPX(); LDY();		  }break; // 4 LDY ZPX
	case 0xD4 : { opcode0xD4();                             	  }break;
	case 0xF4 : { opcode0xF4();                     		  }break;
	    
	case 0x05 : { tmp = 0;	iCount -= 3; RD_ZPG(); ORA();		  }break; // 3 ORA ZPG
	case 0x25 : { tmp = 0;	iCount -= 3; RD_ZPG(); AND();		  }break; // 3 AND ZPG
	case 0x45 : { tmp = 0;	iCount -= 3; RD_ZPG(); EOR();		  }break; // 3 EOR ZPG
	case 0x65 : { tmp = 0;	iCount -= 3; RD_ZPG(); ADC();		  }break; // 3 ADC ZPG
	case 0x85 : { tmp = 0;	iCount -= 3;		 STA(); WR_ZPG(); }break; // 3 STA ZPG
	case 0xA5 : { tmp = 0;	iCount -= 3; RD_ZPG(); LDA();		  }break; // 3 LDA ZPG
	case 0xC5 : { tmp = 0;	iCount -= 3; RD_ZPG(); CMP();		  }break; // 3 CMP ZPG
	case 0xE5 : { tmp = 0;	iCount -= 3; RD_ZPG(); SBC();		  }break; // 3 SBC ZPG
	    
	case 0x15 : { tmp = 0;	iCount -= 4; RD_ZPX(); ORA();		  }break; // 4 ORA ZPX
	case 0x35 : { tmp = 0;	iCount -= 4; RD_ZPX(); AND();		  }break; // 4 AND ZPX
	case 0x55 : { tmp = 0;	iCount -= 4; RD_ZPX(); EOR();		  }break; // 4 EOR ZPX
	case 0x75 : { tmp = 0;	iCount -= 4; RD_ZPX(); ADC();		  }break; // 4 ADC ZPX
	case 0x95 : { tmp = 0;	iCount -= 4;		 STA(); WR_ZPX(); }break; // 4 STA ZPX
	case 0xB5 : { tmp = 0;	iCount -= 4; RD_ZPX(); LDA();		  }break; // 4 LDA ZPX
	case 0xD5 : { tmp = 0;	iCount -= 4; RD_ZPX(); CMP();		  }break; // 4 CMP ZPX
	case 0xF5 : { tmp = 0;	iCount -= 4; RD_ZPX(); SBC();		  }break; // 4 SBC ZPX
	    
	case 0x06 : { tmp = 0;	iCount -= 5; RD_ZPG(); ASL(); WB_EA();  }break; // 5 ASL ZPG
	case 0x26 : { tmp = 0;	iCount -= 5; RD_ZPG(); ROL(); WB_EA();  }break; // 5 ROL ZPG
	case 0x46 : { tmp = 0;	iCount -= 5; RD_ZPG(); LSR(); WB_EA();  }break; // 5 LSR ZPG
	case 0x66 : { tmp = 0;	iCount -= 5; RD_ZPG(); ROR(); WB_EA();  }break; // 5 ROR ZPG
	case 0x86 : { tmp = 0;	iCount -= 3;		 STX(); WR_ZPG(); }break; // 3 STX ZPG
	case 0xA6 : { tmp = 0;	iCount -= 3; RD_ZPG(); LDX();		  }break; // 3 LDX ZPG
	case 0xC6 : { tmp = 0;	iCount -= 5; RD_ZPG(); DEC(); WB_EA();  }break; // 5 DEC ZPG
	case 0xE6 : { tmp = 0;	iCount -= 5; RD_ZPG(); INC(); WB_EA();  }break; // 5 INC ZPG
	    
	case 0x16 : { tmp = 0;	iCount -= 6; RD_ZPX(); ASL(); WB_EA();  }break; // 6 ASL ZPX
	case 0x36 : { tmp = 0;	iCount -= 6; RD_ZPX(); ROL(); WB_EA();  }break; // 6 ROL ZPX
	case 0x56 : { tmp = 0;	iCount -= 6; RD_ZPX(); LSR(); WB_EA();  }break; // 6 LSR ZPX
	case 0x76 : { tmp = 0;	iCount -= 6; RD_ZPX(); ROR(); WB_EA();  }break; // 6 ROR ZPX
	case 0x96 : { tmp = 0;	iCount -= 4;		 STX(); WR_ZPY(); }break; // 4 STX ZPY
	case 0xB6 : { tmp = 0;	iCount -= 4; RD_ZPY(); LDX();		  }break; // 4 LDX ZPY
	case 0xD6 : { tmp = 0;	iCount -= 6; RD_ZPX(); DEC(); WB_EA();  }break; // 6 DEC ZPX
	case 0xF6 : { tmp = 0;	iCount -= 6; RD_ZPX(); INC(); WB_EA();  }break; // 6 INC ZPX
	    
	case 0x07 : { opcode0x07();                              	  }break; 
	case 0x27 : { opcode0x27();                     		  }break; 
	case 0x47 : { opcode0x47();                             	  }break;
	case 0x67 : { opcode0x67();                     		  }break;
	case 0x87 : { opcode0x87();                     		  }break;
	case 0xA7 : { opcode0xA7();                      		  }break;
	case 0xC7 : { opcode0xC7();                       		  }break;
	case 0xE7 : { opcode0xE7();                              	  }break; 
	    
	case 0x17 : { opcode0x17();                     		  }break;
	case 0x37 : { opcode0x37();                     		  }break; 
	case 0x57 : { opcode0x57();                      		  }break;
	case 0x77 : { opcode0x77();                     		  }break;
	case 0x97 : { opcode0x97();                     		  }break; 
	case 0xB7 : { opcode0xB7();                     		  }break;
	case 0xD7 : { opcode0xD7();                     		  }break;
	case 0xF7 : { opcode0xF7();                      		  }break;
	    
	case 0x08 : {		  iCount -= 2;		 PHP();		  }break; // 2 PHP
	case 0x28 : {		  iCount -= 2;		 PLP();		  }break; // 2 PLP
	case 0x48 : {		  iCount -= 2;		 PHA();		  }break; // 2 PHA
	case 0x68 : {		  iCount -= 2;		 PLA();		  }break; // 2 PLA
	case 0x88 : {		  iCount -= 2;		 DEY();		  }break; // 2 DEY
	case 0xA8 : {		  iCount -= 2;		 TAY();		  }break; // 2 TAY
	case 0xC8 : {		  iCount -= 2;		 INY();		  }break; // 2 INY
	case 0xE8 : {		  iCount -= 2;		 INX();		  }break; // 2 INX
	    
	case 0x18 : {		  iCount -= 2;		 CLC();		  }break; // 2 CLC
	case 0x38 : {		  iCount -= 2;		 SEC();		  }break; // 2 SEC
	case 0x58 : {		  iCount -= 2;		 CLI();		  }break; // 2 CLI
	case 0x78 : {		  iCount -= 2;		 SEI();		  }break; // 2 SEI
	case 0x98 : {		  iCount -= 2;		 TYA();		  }break; // 2 TYA
	case 0xB8 : {		  iCount -= 2;		 CLV();		  }break; // 2 CLV
	case 0xD8 : {		  iCount -= 2;		 CLD();		  }break; // 2 CLD
	case 0xF8 : {		  iCount -= 2;		 SED();		  }break; // 2 SED

	case 0x09 : { tmp = 0; iCount -= 2; RD_IMM(); ORA();		  }break; // 2 ORA IMM
	case 0x29 : { tmp = 0; iCount -= 2; RD_IMM(); AND();		  }break; // 2 AND IMM
	case 0x49 : { tmp = 0; iCount -= 2; RD_IMM(); EOR();		  }break; // 2 EOR IMM
	case 0x69 : { tmp = 0; iCount -= 2; RD_IMM(); ADC();		  }break; // 2 ADC IMM
	case 0x89 : { opcode0x89();                     		  }break;
	case 0xA9 : { tmp = 0; iCount -= 2; RD_IMM(); LDA();		  }break; // 2 LDA IMM
	case 0xC9 : { tmp = 0; iCount -= 2; RD_IMM(); CMP();		  }break; // 2 CMP IMM
	case 0xE9 : { tmp = 0; iCount -= 2; RD_IMM(); SBC();		  }break; // 2 SBC IMM

	case 0x19 : { tmp = 0; iCount -= 4; RD_ABY(); ORA();		  }break; // 4 ORA ABY
	case 0x39 : { tmp = 0; iCount -= 4; RD_ABY(); AND();		  }break; // 4 AND ABY
	case 0x59 : { tmp = 0; iCount -= 4; RD_ABY(); EOR();		  }break; // 4 EOR ABY
	case 0x79 : { tmp = 0; iCount -= 4; RD_ABY(); ADC();		  }break; // 4 ADC ABY
	case 0x99 : { tmp = 0; iCount -= 5;		 STA(); WR_ABY(); }break; // 5 STA ABY
	case 0xB9 : { tmp = 0; iCount -= 4; RD_ABY(); LDA();		  }break; // 4 LDA ABY
	case 0xD9 : { tmp = 0; iCount -= 4; RD_ABY(); CMP();		  }break; // 4 CMP ABY
	case 0xF9 : { tmp = 0; iCount -= 4; RD_ABY(); SBC();		  }break; // 4 SBC ABY

	case 0x0A : { tmp = 0; iCount -= 2; RD_ACC(); ASL(); WB_ACC(); }break; // 2 ASL A
	case 0x2A : { tmp = 0; iCount -= 2; RD_ACC(); ROL(); WB_ACC(); }break; // 2 ROL A
	case 0x4A : { tmp = 0; iCount -= 2; RD_ACC(); LSR(); WB_ACC(); }break; // 2 LSR A
	case 0x6A : { tmp = 0; iCount -= 2; RD_ACC(); ROR(); WB_ACC(); }break; // 2 ROR A
	case 0x8A : {		  iCount -= 2;		 TXA();		  }break; // 2 TXA
	case 0xAA : {		  iCount -= 2;		 TAX();		  }break; // 2 TAX
	case 0xCA : {		  iCount -= 2;		 DEX();		  }break; // 2 DEX
	case 0xEA : {		  iCount -= 2;		 NOP();		  }break; // 2 NOP

	case 0x1A : { opcode0x1A();                     		  }break; 
	case 0x3A : { opcode0x3A();                             	  }break;
	case 0x5A : { opcode0x5A();                     		  }break;
	case 0x7A : { opcode0x7A();                     		  }break;
	case 0x9A : {		  iCount -= 2;		 TXS();		  }break; // 2 TXS
	case 0xBA : {		  iCount -= 2;		 TSX();		  }break; // 2 TSX
	case 0xDA : { opcode0xDA();                     		  }break;
	case 0xFA : { opcode0xFA();                     		  }break;

	case 0x0B : { opcode0x0B();                     		  }break; 
	case 0x2B : { opcode0x2B();                        		  }break;
	case 0x4B : { opcode0x4B();                     		  }break;
	case 0x6B : { opcode0x6B();                     		  }break;
	case 0x8B : { opcode0x8B();                     		  }break;
	case 0xAB : { opcode0xAB();                     		  }break;
	case 0xCB : { opcode0xCB();                     		  }break;
	case 0xEB : { opcode0xEB();                     		  }break;

	case 0x1B : { opcode0x1B();                     		  }break;
	case 0x3B : { opcode0x3B();                     		  }break;
	case 0x5B : { opcode0x5B();                      		  }break;
	case 0x7B : { opcode0x7B();                     		  }break;
	case 0x9B : { opcode0x9B();                     		  }break;
	case 0xBB : { opcode0xBB();                     		  }break;
	case 0xDB : { opcode0xDB();                     		  }break;
	case 0xFB : { opcode0xFB();                     		  }break;

	case 0x0C : { opcode0x0C();                     		  }break;
	case 0x2C : { tmp = 0; iCount -= 4; RD_ABS(); BIT();		  }break; // 4 BIT ABS
	case 0x4C : {		  iCount -= 3; EA_ABS(); JMP();		  }break; // 3 JMP ABS
	case 0x6C : { tmp = 0; iCount -= 5; EA_IND(); JMP();		  }break; // 5 JMP IND
	case 0x8C : { tmp = 0; iCount -= 4;		 STY(); WR_ABS(); }break; // 4 STY ABS
	case 0xAC : { tmp = 0; iCount -= 4; RD_ABS(); LDY();		  }break; // 4 LDY ABS
	case 0xCC : { tmp = 0; iCount -= 4; RD_ABS(); CPY();		  }break; // 4 CPY ABS
	case 0xEC : { tmp = 0; iCount -= 4; RD_ABS(); CPX();		  }break; // 4 CPX ABS

	case 0x1C : { opcode0x1C();                     		  }break;
	case 0x3C : { opcode0x3C();                     		  }break;
	case 0x5C : { opcode0x5C();                     		  }break;
	case 0x7C : { opcode0x7C();                      		  }break;
	case 0x9C : { opcode0x9C();                     		  }break;
	case 0xBC : { tmp = 0; iCount -= 4; RD_ABX(); LDY();		  }break; // 4 LDY ABX
	case 0xDC : { opcode0xDC();                     		  }break;
	case 0xFC : { opcode0xFC();                     		  }break;

	case 0x0D : { tmp = 0; iCount -= 4; RD_ABS(); ORA();		  }break; // 4 ORA ABS
	case 0x2D : { tmp = 0; iCount -= 4; RD_ABS(); AND();		  }break; // 4 AND ABS
	case 0x4D : { tmp = 0; iCount -= 4; RD_ABS(); EOR();		  }break; // 4 EOR ABS
	case 0x6D : { tmp = 0; iCount -= 4; RD_ABS(); ADC();		  }break; // 4 ADC ABS
	case 0x8D : { tmp = 0; iCount -= 4;		 STA(); WR_ABS(); }break; // 4 STA ABS
	case 0xAD : { tmp = 0; iCount -= 4; RD_ABS(); LDA();		  }break; // 4 LDA ABS
	case 0xCD : { tmp = 0; iCount -= 4; RD_ABS(); CMP();		  }break; // 4 CMP ABS
	case 0xED : { tmp = 0; iCount -= 4; RD_ABS(); SBC();		  }break; // 4 SBC ABS

	case 0x1D : { tmp = 0; iCount -= 4; RD_ABX(); ORA();		  }break; // 4 ORA ABX
	case 0x3D : { tmp = 0; iCount -= 4; RD_ABX(); AND();		  }break; // 4 AND ABX
	case 0x5D : { tmp = 0; iCount -= 4; RD_ABX(); EOR();		  }break; // 4 EOR ABX
	case 0x7D : { tmp = 0; iCount -= 4; RD_ABX(); ADC();		  }break; // 4 ADC ABX
	case 0x9D : { tmp = 0; iCount -= 5;		 STA(); WR_ABX(); }break; // 5 STA ABX
	case 0xBD : { tmp = 0; iCount -= 4; RD_ABX(); LDA();		  }break; // 4 LDA ABX
	case 0xDD : { tmp = 0; iCount -= 4; RD_ABX(); CMP();		  }break; // 4 CMP ABX
	case 0xFD : { tmp = 0; iCount -= 4; RD_ABX(); SBC();		  }break; // 4 SBC ABX

	case 0x0E : { tmp = 0; iCount -= 6; RD_ABS(); ASL(); WB_EA();  }break; // 6 ASL ABS
	case 0x2E : { tmp = 0; iCount -= 6; RD_ABS(); ROL(); WB_EA();  }break; // 6 ROL ABS
	case 0x4E : { tmp = 0; iCount -= 6; RD_ABS(); LSR(); WB_EA();  }break; // 6 LSR ABS
	case 0x6E : { tmp = 0; iCount -= 6; RD_ABS(); ROR(); WB_EA();  }break; // 6 ROR ABS
	case 0x8E : { tmp = 0; iCount -= 5;		 STX(); WR_ABS(); }break; // 5 STX ABS
	case 0xAE : { tmp = 0; iCount -= 4; RD_ABS(); LDX();		  }break; // 4 LDX ABS
	case 0xCE : { tmp = 0; iCount -= 6; RD_ABS(); DEC(); WB_EA();  }break; // 6 DEC ABS
	case 0xEE : { tmp = 0; iCount -= 6; RD_ABS(); INC(); WB_EA();  }break; // 6 INC ABS

	case 0x1E : { tmp = 0; iCount -= 7; RD_ABX(); ASL(); WB_EA();  }break; // 7 ASL ABX
	case 0x3E : { tmp = 0; iCount -= 7; RD_ABX(); ROL(); WB_EA();  }break; // 7 ROL ABX
	case 0x5E : { tmp = 0; iCount -= 7; RD_ABX(); LSR(); WB_EA();  }break; // 7 LSR ABX
	case 0x7E : { tmp = 0; iCount -= 7; RD_ABX(); ROR(); WB_EA();  }break; // 7 ROR ABX
	case 0x9E : { opcode0x9E();                   			   }break; // 2 ILL
	case 0xBE : { tmp = 0; iCount -= 4; RD_ABY(); LDX();		  }break; // 4 LDX ABY
	case 0xDE : { tmp = 0; iCount -= 7; RD_ABX(); DEC(); WB_EA();  }break; // 7 DEC ABX
	case 0xFE : { tmp = 0; iCount -= 7; RD_ABX(); INC(); WB_EA();  }break; // 7 INC ABX

	case 0x0F : { opcode0x0F();                     		  }break;
	case 0x2F : { opcode0x2F();                     		  }break;
	case 0x4F : { opcode0x4F();                     		  }break;
	case 0x6F : { opcode0x6F();                     		  }break;
	case 0x8F : { opcode0x8F();                      		  }break;
	case 0xAF : { opcode0xAF();                     		  }break;
	case 0xCF : { opcode0xCF();                       		  }break;
	case 0xEF : { opcode0xEF();                     		  }break;

	case 0x1F : { opcode0x1F();                     		  }break;
	case 0x3F : { opcode0x3F();                      		  }break;
	case 0x5F : { opcode0x5F();                        		  }break;
	case 0x7F : { opcode0x7F();                     		  }break;
	case 0x9F : { opcode0x9F();                      		  }break;
	case 0xBF : { opcode0xBF();                     		  }break;
	case 0xDF : { opcode0xDF();                     		  }break; 
	case 0xFF : { opcode0xFF();                         		  } 
	}
    }

    /** Reset the processor. */
    final protected void reset(){
	/* wipe out the rest of the m6502 structure */
	/* read the reset vector into PC */
	programCounter.setLowByte( RDMEM( M6502_RST_VEC ) );
	programCounter.setHighByte( RDMEM( M6502_RST_VEC + 1 ) );
	stackPointer.setDWord( 0x01FF );	/* stack pointer starts at 0x01FF */
	processorStatus.set( F_T | F_I | F_Z );	/* set T, I and Z flags */
	pending_irq = 0; /* nonzero if an IRQ is pending */
	after_cli = 0; /* pending IRQ and last insn cleared I */
	//m6502.irq_callback = NULL;
	changePC16( programCounter.getDWord() );
    }

    /** Main loop.
	@param cycles the number of cycles available to execute opcodes.
	@return the number of extra cycles (i.e. unavailable cycles) consumed by the last opcode .*/
    final protected int execute( int cycles ){
	iCount = cycles;
	changePC16( programCounter.getDWord() );
	do
	    {
		final UInt8 op = new UInt8();
		previousProgramCounter.setDWord( programCounter.getDWord() );
		op.set( readOpcode() );
		/* if an irq is pending, take it now */
		if( ( pending_irq != 0 ) && ( op.get() == 0x78 ) )
		    {
			take_irq();
		    }
		switchOpcode( op.get() );
		/* check if the I flag was just reset (interrupts enabled) */
		if ( after_cli != 0 )
		    {
			after_cli = 0;
			if ( irq_state != CLEAR_LINE )
			    {
				pending_irq = 1;
			    }
		    }
		else
		    {
			if ( pending_irq != 0 )
			    {
				take_irq();
			    }
		    }
	    }
	while ( iCount > 0 );
	return( cycles - iCount );
    }

    /** Set NMI line.
	@param state is the NMI line state.<BR>
	If state is 0, the line is inactive, else the line is active. May be revised to become a boolean? */
    final protected void set_nmi_line( int state ){
	if ( nmi_state == state )
	    {
		return;
	    }
	nmi_state = state;
	if( state != CLEAR_LINE )
	    {
		effectiveAddress.setDWord( M6502_NMI_VEC );
		iCount -= 7;
		PUSH( programCounter.getHighByte() );
		PUSH( programCounter.getLowByte() );
		PUSH( processorStatus.get() & ~F_B );
		processorStatus.set( ( processorStatus.get() & ~F_D ) | F_I ); /* knock out D and set I flag */
		programCounter.setLowByte( RDMEM( effectiveAddress.getDWord() ) );
		programCounter.setHighByte( RDMEM( effectiveAddress.getDWord() + 1 ) );
		changePC16( programCounter.getDWord() );
	    }
    }

    /** Set IRQ line.
	@param irqline is the IRQ line.<BR>
	@param state is the IRQ line state.<BR>
	If state is 0, the line is inactive, else the line is active. May be revised to become a boolean? */
    final protected void set_irq_line( int irqline, int state ){
	if( irqline == M6502_SET_OVERFLOW )
	    {
		if( ( so_state != 0 ) && !( state != 0 ) )
		    {
			processorStatus.set( processorStatus.get() | F_V );
		    }
		so_state = state;
		return;
	    }
	irq_state = state;
	if( state != CLEAR_LINE )
	    {
		pending_irq = 1;
	    }
    }
    
    /** Set IRQ callback.
	@param callback ???.<BR>
	Not implemented yet. */
    final private void m6502_set_irq_callback( int callback )
    {
	irq_callback = callback;
    }
    

    /** Take IRQ. */
    final protected void take_irq(){
	if( !( ( processorStatus.get() & F_I ) != 0 ) ){
	    effectiveAddress.setDWord( M6502_IRQ_VEC  );
	    iCount -= 7;
	    PUSH( programCounter.getHighByte() );
	    PUSH( programCounter.getLowByte() );
	    PUSH( processorStatus.get() & ~F_B );
	    processorStatus.set( ( processorStatus.get() & ~F_D ) | F_I ); /* knock out D and set I flag */
	    programCounter.setLowByte( RDMEM( effectiveAddress.getDWord() ) );
	    programCounter.setHighByte( RDMEM( effectiveAddress.getDWord() + 1 ) );
	    /* call back the cpuintrf to let it clear the line */
	    //if ( m6502.irq_callback ) (*m6502.irq_callback)(0); //pending
	    changePC16( programCounter.getDWord() );
	}
	pending_irq = 0;
    }

}
